/*->c.spell    */

/* Ovation!   (c) D. J. Pilling,  December 1988                     */
/*                        Spelling Checker                            */

#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>
#include <time.h>

#include "os.h"
#include "wimp.h"
#include "sprite.h"
#include "bbc.h"
#include "font.h"
#include "werr.h"

#include "fsx.h"
#include "bf.h"
#include "bits.h"
#include "err.h"
#include "conf.h"
#include "hourglass.h"
#include "wos.h"
#include "deb.h"
#include "poll.h"
#include "etc.h"


#include "main.h"
#include "soln.h"
#include "grid.h"
#include "deb.h"
#include "memblock.h"
#include "dc.h"
#include "edit.h"
#include "reslink.h"
#include "mainwin.h"
#include "file.h"



#include "spell.h"




/*****************************************************************************/
                            /* Edit a dictionary */



dline dtable[MAXNAMES];

int   totnames=0;                 /* number of dictionaries     */

char  dcmap[MAXNAMES];            /* dictionaries for file      */
int   nodc;

#define DMSIZE 0x80

int   dictm[DMSIZE];






/*****************************************************************************/





void savechanges(void)
{
 int  i;

 for(i=0;i<totnames;i++) 
 {
  if(dtable[i].modded)
  {
   cprintf("savechanges");
   savedc(dictionarydir,dtable[i].name,&dtable[i].dictp);
   dtable[i].modded=0;
  }
 }
}




void scrapchanges(void)
{
 int i;
 for(i=0;i<totnames;i++) 
          if(dtable[i].modded)
           {
            trashdc(&dtable[i].dictp);
            loaddcn(i);
            dtable[i].modded=0;
           }
}




int changes(void)
{
 int i;

 

 for(i=0;i<totnames;i++)
 {
  cprintf("changes %d",dtable[i].modded);
  if(dtable[i].modded) return(1);
 }
 return(0);
}




/*****************************************************************************/



void initline(int line)
{
 int i;

 for(i=0;i<12;i++) dtable[line].name[i]=0;

 dtable[line].dictp.nowords=0;
 dtable[line].dictp.noindex=0;
 dtable[line].dictp.words=NULL;
 dtable[line].dictp.index=NULL;
 dtable[line].loaded=0;
 dtable[line].uses=0;
 dtable[line].modded=0;
}



void inittable(void)
{
 int line;
 for(line=0;line<MAXNAMES;line++) initline(line);
 totnames=0;
}




/* loads the dictionary names */

os_error * loadnames(void)
{
 os_error * err;
 fxstat     f;
 int        eof;
 char     * p;

 /* scan the dictionary directory */

 err=NULL;

 cprintf("dic dir %s",dictionarydir);

 totnames=0;
 fs_startscan();
 while(1)
 { 
  err=fs_nextitem(dictionarydir,&f,"*.dic",&eof);
  if(err) break;
  if(eof) break;

  cprintf("name %s",f.name);

  p=fs_extension(f.name);
  *(p-1)=0;

  strcpy(dtable[totnames].name,f.name);
  totnames++;
  if(totnames>=MAXNAMES) break;
 }
 fs_endscan();

 return(err);
}


/* given a dictionary number - load the dictionary */

os_error * loaddcn(int n)
{
 os_error * err;

 err=loaddc(dictionarydir,dtable[n].name,&dtable[n].dictp);

 if(err) dtable[n].loaded=0;
 else    dtable[n].loaded=1;

 return(err);
}






/* given a dictionary number - remove the dictionary from memory */

int unloaddcn(int n)
{
 trashdc(&dtable[n].dictp);
 dtable[n].loaded=0;
 return(1);
}




void dccleanup(void)
{
 int i;
 for(i=0;i<totnames;i++) 
          if(!dtable[i].uses && dtable[i].loaded) unloaddcn(i);
}



/* given a dictionary name, return its number or -1 */

int finddc(char * name)
{
 int i=0;
 while(i<totnames) if(!cstrcmp(dtable[i].name,name)) return(i); else i++;
 return(-1);
}


/* how many dictionaries are there? */

int nodcs(void)
{
 return(totnames);
}


/* how many dictionaries are there excluding hyphenation dictionary */

int nodcsexhp(void)
{
 return(totnames);
}





/*****************************************************************************/
           /* Edit dictionary menu - load,edit,create,delete */


os_error * dcadd(int n)
{
 os_error * err;
 int        i;
 int        j;

 err=NULL;

 for(i=0;i<nodc;i++) if(dcmap[i]>=n) break;

 if(nodc==MAXNAMES) err=geterror(EMDC); // Cannot load anymore dictionaries for this document.
 else
 {
  for(j=nodc+1;j>i;j--) dcmap[j]=dcmap[j-1];

  dcmap[i]=(char)n;
  nodc++;
  dtable[n].uses++;
 }

 return(err);
}


os_error * dcaddload(int n)
{
 os_error * err;

 err=NULL;
 
 dcadd(n);
 if(!dtable[n].loaded) loaddcn(n);

 return(err);
}



/* remove a dictionary from a file */

static os_error * dcrem(int n)
{
 int i;
 int j;

 for(i=0;i<nodc;i++) if(dcmap[i]>=n) break;

 if(nodc && i<nodc && dcmap[i]==n)
 {
  for(j=i;j<(nodc-1);j++) dcmap[j]=dcmap[j+1];

  nodc--;
  dtable[n].uses--;

  if(!dtable[n].uses && dtable[n].loaded) unloaddcn(n);
 }

 return(NULL);
}


/* does file use this dictionary ? */

int dcfile(int n)
{
 int i;
 for(i=0;i<nodc;i++) if(dcmap[i]==n) return(1);
 return(0);
}



/* add the dictionary to the current document */

os_error * loaddict(int m3)
{
 os_error * err;
 
 if(dcfile(m3)) err=dcrem(m3);
 else           err=dcaddload(m3);

 modfile();

 return(err);
}




os_error * editdict(int m3)
{
 os_error * err;
 
 if(!dtable[m3].loaded) err=loaddcn(m3);
 else                   err=NULL;

 if(!err)
 {
  openeditd(mainhandle,m3,&dtable[m3].dictp,dtable[m3].name);
 }


 return(err);
}



void dcins(int n)
{
 int j;

 for(j=0;j<nodc;j++) if(dcmap[j]>=n) dcmap[j]++;
}



/* called from menu to create a dictionary */

os_error * createdict(char * string)
{
 os_error * err;
 int        i;
 int        n;

 err=NULL;

 if(totnames==MAXNAMES) err=geterror(ENUD); // Cannot create anymore user dictionaries.
 {
  for(i=0;i<totnames;i++)
  {
   if(!cstrcmp(dtable[i].name,string))
   {
    err=geterror(EDAX); // Dictionary already exists.
    break;
   }
  }

  if(!err)
  {
   for(n=0;n<totnames;n++)
   {
    if(strcmp(dtable[n].name,string)>0) break;
   }


   totnames++;
   for(i=totnames-1;i>n;i--) 
   {
    dtable[i]=dtable[i-1];
    relocmemblock(&dtable[i-1].dictp.words,&dtable[i].dictp.words);
    relocmemblock(&dtable[i-1].dictp.index,&dtable[i].dictp.index);
   }

   initline(n);
   strcpy(dtable[n].name,string);
   dtable[n].loaded=1;

   dcins(n);

   createdc(&dtable[n].dictp);
   err=savedc(dictionarydir,dtable[n].name,&dtable[n].dictp);
   if(err)
   {
    deletedictsub(n);
   }
  }

  dcrebuild();
 }

 return(err);
}



void dcdel(int n)
{
 int j;
 dcrem(n);
 for(j=0;j<nodc;j++) if(dcmap[j]>n) dcmap[j]--;
}


os_error * deletedictsub(int m3)
{
 os_error * err;
 char dictionarypath[MAX_PATH];
 int  i;

 sprintf(dictionarypath,"%s\\%s.dic",dictionarydir,dtable[m3].name);

 err=fs_delete(dictionarypath);
 if(!err)
 {
  dcdel(m3);

  for(i=m3;i<(totnames-1);i++) 
  {
   dtable[i]=dtable[i+1];
   relocmemblock(&dtable[i+1].dictp.words,&dtable[i].dictp.words);
   relocmemblock(&dtable[i+1].dictp.index,&dtable[i].dictp.index);
  }

  totnames--;
 }

 return(err);
}



os_error * deletedict(int m3)
{
 os_error * err;
 
 // if(confirm("Deleting Dictionary, Are You Sure?")!=1) return;  
 err=deletedictsub(m3);
 if(!err)
 {
  dcrebuild();
 }

 return(err);
}



/*************************************************************************/
             /* handle dictionary loading within documents */



/* return 1 if any errors */
int readfhdc(FILE * fp)
{
 char name[12];
 int  dcn;

 nodc=0;

 while(1)
 {
  fread(name,12,1,fp);
  if(ferror(fp)) return(1);
  if(!strlen(name)) break;
  dcn=finddc(name);
  if(dcn>-1) dcadd(dcn);
 }

 dccleanup();

 return(0);
}


/* return 1 if any errors */

int writefhdc(FILE * fp)
{
 int i;
 char name[12];

 for(i=0;i<nodc;i++) 
          fwrite(dtable[dcmap[i]].name,12,1,fp);

 for(i=0;i<12;i++) name[i]=0;
 fwrite(name,12,1,fp);

 if(ferror(fp)) return(1);
 return(0);
}


void mapfhdc(void)
{
 int i;

 for(i=0;i<nodc;i++) 
     if(!dtable[dcmap[i]].loaded) loaddcn(dcmap[i]);
}



void newfhdcmap(void)
{
 nodc=0;
}



void trashfhdc(void)
{
 int i;
 for(i=0;i<nodc;i++) dtable[dcmap[i]].uses--;
 nodc=0;
}


/*****************************************************************************/





char * swordn(int word)
{
 int dc=word>>24;
 char * w=(wordn(word & 0xFFFFFF,&dtable[dc].dictp));
 return(w);
}






#define MSPLIT 3


void missingletter(char * string,dict * dp,word * words)
{
 char * c;
 int    n;
 int    match=32;
 int    miss1;
 int    miss2;
 int    offset;
 char   work[32];
 char   * w;
 char   * s;
 int      m;

 offset=words->woffset+words->len;
 strcpy(work,words->string);
 c=charblock(dp->words,offset);

 n=*c;
 while(1)
 {
  work[n]=0;

  if(n>(match+1))
  {
   while(1)
   {
    c++;
    if((n=*c)<MINTOK)
    {
     if(n<ESCTOK) break;
    }
   }
  }
  else
  {
   while(1)
   {
    c++;
    if((n=*c)<MINTOK)
    {
     if(n<ESCTOK) break;
     else
     {
      m=strlen(work);
      c++;
      work[m++]=*c;
      work[m]=0;
     }

    }
    else
    {
     strcat(work,mfx[n]);
    }
   }

   match=miss1=miss2=0;
   w=work;
   s=string;

   while(1)
   {
    if(*w!=*s && toupper(*w)!=toupper(*s))
    {
/*     if(!*s) break;
     if(!*w) break;  */
  
     if(miss1 && miss2) break;
     else
     if(!miss1 && *w && *(w+1)==*s) {miss1=1;w++;match++;}
     else
     if(!miss2 && *s && *w==*(s+1)) {miss2=1;s++;}
     else
     if(!miss1 && !miss2) {miss1=miss2=1;w++;s++;match++;}
     else
     break;
    }
    else
    if(!*w)
    {
     offset=c-charblock(dp->words,0);
     addword(work,&guessdc);
     c=charblock(dp->words,offset);
     break;
    }
    else
    {
     match++;
     w++;
     s++;
    }
   }

  }

  if(match<MSPLIT) break;
 }

}


void makeguess2(char * string)
{
 char temp[64];
 word words;
 int  len=strlen(string);
 int  lenx;
 int  i;
 int  j;
 int  k;
 int  flags=0;

 /* closest in dictionary */

 for(i=0;i<nodc;i++)
 {
  findword(string,&dtable[dcmap[i]].dictp,&words,&flags);
  if(toupper(string[0])==toupper(words.string[0]))
                                         addword(words.string,&guessdc);
  else
  {
   nextword(&dtable[dcmap[i]].dictp,&words);
   if(toupper(string[0])==toupper(words.string[0]))
                                         addword(words.string,&guessdc);
  }
 }


 /* permutations of letters */

 flags=0x20000;

 for(j=0;j<len-1;j++)
 {
  strcpy(temp,string);
  temp[j]=temp[j+1];
  temp[j+1]=string[j];

  for(i=0;i<nodc;i++)
    if(findword(temp,&dtable[dcmap[i]].dictp,&words,&flags))
     {
      xgetscanstring(words.string);
      addword(words.string,&guessdc);
      break;
     }
 }


 /* missing letters out */

 if(len>1)
   for(j=0;j<len;j++)
   {
    for(i=0;i<=len;i++) temp[i-(i>j)]=string[i];

    for(i=0;i<nodc;i++)
     if(findword(temp,&dtable[dcmap[i]].dictp,&words,&flags))
     {
      xgetscanstring(words.string);
      addword(words.string,&guessdc);
      break;
     }
   }


 /* adding extra letters */

 if(len>MSPLIT) lenx=MSPLIT; else lenx=len;

 for(k=0;k<=lenx;k++)
 {
  for(j=0;j<=len;j++) temp[j+(j>=k)]=string[j];

  for(j='a';j<='z';j++)
  {
   temp[k]=(char)j;
   for(i=0;i<nodc;i++)
   if(findword(temp,&dtable[dcmap[i]].dictp,&words,&flags))
   {
    xgetscanstring(words.string);
    addword(words.string,&guessdc);
   }
  }
 }


 /* wrong letters */

 for(k=1;k<=lenx;k++)
 {
  strcpy(temp,string);

  for(j='a';j<='z';j++)
  {
   temp[k]=(char)j;
   for(i=0;i<nodc;i++)
   if(findword(temp,&dtable[dcmap[i]].dictp,&words,&flags))
   {
    xgetscanstring(words.string);
    addword(words.string,&guessdc);
   }
  }
 }



 flags=0;

 if(len>MSPLIT)
 {
  strcpy(temp,string);
  temp[MSPLIT]=0;

  for(i=0;i<nodc;i++)
  {
   findword(temp,&dtable[dcmap[i]].dictp,&words,&flags);
   missingletter(string,&dtable[dcmap[i]].dictp,&words);
  }
 }

}


/* make guesses for word */

void makeguess(char * string)
{
 strcpy(guessit,string);
 trashdc(&guessdc);
 createdc(&guessdc);
 makeguess2(string);
}

/*****************************************************************************/





/*

The rules for capitalization are as follows:

Any word may appear in all capitals, as in headings.

Any word that is in the dictionary in all-lowercase form may appear
either in lowercase or capitalized (as at the beginning of a sentence).

Any word that has "funny" capitalization (i.e., it contains both cases
and there is an uppercase character besides the first) must appear
exactly as in the dictionary, except as permitted by rule (1).
If the word is acceptable in all-lowercase, it must appear thus in a
dictionary entry.

*/


int spellcheckwordsub(char * string)
{
 word words;
 int  i;
 int  flags=0x20000;


 for(i=0;i<nodc;i++)
 {
  if(findword(string,&dtable[dcmap[i]].dictp,&words,&flags))
  {
 //  if(!(flags & 0x1)) return(1);  /* case matched */
   return(1);
  }
 }
 return(0);
}





/* wrapped up spell check, handles multiple file dictionaries */

int spellcheckword(char * string1)
{
 return(spellcheckwordsub(string1));
}

int cluenumber;


/* return 2 word OK - close check, 1 word OK or 0 error */

int checknextsub(int oneshot,int checker)
{
 char   string[64];
 int    i;

 while(nextclue(string,cluenumber++))
 {
  i=0;
  while(string[i])
  {
   if(string[i]==32) break;
   else i++;
  }

  if(!string[i])   /* don't check words with spaces in */
  {
   strcpy(checkstring,string);

   if(!spellcheckword(string)) 
   {
    if(checker)
    {
     if(!spellerror(string)) return(2);
    }
    return(0);
   }
  }
 }

 return(2);

 USE(oneshot);
}


int checknext(int oneshot,int checker)
{
 int code;
 hourglasson();
 code=checknextsub(oneshot,checker);
 hourglassoff();

 cprintf("check next %d",code);
 if(code==2 || code==3) 
 {
  if(checkmode==CHECKS) closecheck(1);
  else
  if(code==2) messagebox("No new words!");
  code=1;
 }
 return(code);
}




void checkcont(void)
{
 checknext(0,1);
}


static char * defltdcnames[]=
{
 "MainDict",
 "MainDict-UK",
 "MainDict-US",
 "Puzzle-UK",
 "Puzzle-US",

};

#define DCNAMEN (sizeof(defltdcnames)/sizeof(char*))

os_error * checkfiledc(void)
{
 os_error * err;
 int        dcno;
 int        i;

 err=NULL;

 if(!nodc)
 {
  for(i=0;i<DCNAMEN;i++)
  {
   dcno=finddc(defltdcnames[i]);
   if(dcno!=-1)
   {
    dcaddload(dcno);
    break;
   }
  }
 }

 if(!nodc)
 {
  err=geterror(ENOD); // No dictionaries loaded!
 }

 return(err);
}



/* check current source */

os_error * spellcheck(void)
{
 cluenumber=0;
 cprintf("spell check");
 if(checkfiledc()) return(NULL);
 checknext(0,1);
 return(NULL);
}




/****************************************************************************/

static int remspacecopy(char * d,char * s)
{
 char * p=d;
 int    c;

 while(1)
 {
  c=*s++;
  if(c!=' ') *d++=c;
  if(!c) break;
 }

 return(d-p-1);
}


int findwordx1(char * string,int dc,word * words,int firstspace,int len,int clue)
{
 int         n;
 int         lo;
 int         hi;
 int         i;
 int         loi;
 int         hii;
 int         ii;
 int         code;
 index    *  f;
 char     *  p;
 int         flags=DCIGNORESPACE;
 dict     *  dp;
 int         check;
 char        wstring[STRLEN];
 int         wlen;


 dp=&dtable[dcmap[dc]].dictp;
 n=dp->noindex;

 lo=0;
 hi=(n-1)*sizeof(index);
 i=0;
 loi=0;
 hii=n-1;
 ii=0;
 check=0;

 if(!dp->nowords)
 {
  words->woffset=0;
  words->indexn=0;
  strcpy(words->string,"");
  words->wordn=0;
  words->len=0;
 }
 else
 {
  p=charblock(dp->index,0);

  if(firstspace>KEYLEN) firstspace=KEYLEN;

  while(hi>=lo)
  {
   i=(lo+hi)>>1;
   ii=loi+hii;
   if((ii & 1) && i) i-=(sizeof(index)>>1);
   ii=ii>>1;

   f=(index *)(p+i);

   code=xstrncmp(string,f->key,firstspace,&flags);

   if(!code) break;
   else
   if(code>0) {lo=i+sizeof(index);loi=ii+1;}
   else       {hi=i-sizeof(index);hii=ii-1;}
  }

  while((code<=0) && (i>0))
  {
   i-=sizeof(index);
   ii--;
   f=(index *)(p+i);
   code=xstrncmp(string,f->key,firstspace,&flags);
  }

  strncpy(words->string,f->key,KEYLEN);
  words->woffset=WOFF(f);
  words->indexn=ii;
  words->wordn=WORDN(f);
  words->len=0;

  formword(dp,words);

 /* oldwords=*words; */

  while(1)
  {
   wlen=remspacecopy(wstring,words->string);

   if(len==wlen)
   {
    for(i=0;i<len;i++)
    {
     if(string[i]!=' ' && tolower(wstring[i])!=tolower(string[i])) break;
     else
     if(wstring[i] && !string[i]) break;
    }
 
    if(i==len && !string[i])
    {
     cluestr[clue].dcdata=(dcmap[dc]<<24)+words->wordn;
  
     if(clue==(solwidth-1))
     {
      for(i=clue;i>=0;i--)
      {
       addsolution(cluestr[i].dcdata,i);
      }
      if(nsol>solnmaxn)
      {
       check=-1;
       break;
      }
     }
     else
     {
      fillinstring(clue,wstring);
      check=search(clue+1);
      fillinstring(clue,string);
     }
    }
    else
    if(i<firstspace)
    {
     code=xstrncmp(string,words->string,firstspace,&flags);
     if(code<0) break;
    }
   }

/* oldwords=*words; */

   if(!nextword(dp,words)) break;

   if((check>=0) && (check++==2048))
   {
    check=0;
    if(buttonsevent())
    {
     check=-1;
     break;
    }
   }
   else
   if(check<0) break;

  }

 /* *words=oldwords; */
 }
 return(check);
}






int search(int clue)
{
 word   words;
 dict * dp;
 int    dc;
 int    len;
 int    i;
 char   string[STRLEN];
 int    firstalpha;
 int    firstspace;
 int    flags=DCIGNORESPACE;
 int    check;


 gridgetstring(clue,string);
 len=strlen(string);
 hourglasson();
 check=0;

 firstalpha=-1;
 firstspace=-1;

 for(i=0;i<len;i++)
 {
  if(isalpha(string[i]) && firstalpha<0) firstalpha=i;
  else
  if(string[i]==32 && firstspace<0)      firstspace=i;
 }

 for(dc=0;dc<nodc;dc++)
 {
  dp=&dtable[dcmap[dc]].dictp;

  if(firstspace==-1)    /* special case, no spaces, can just do a look up */
  {
   if(findword(string,dp,&words,&flags))
   {
    cluestr[clue].dcdata=(dcmap[dc]<<24)+words.wordn;

    if(clue==(solwidth-1))
    {
     for(i=clue;i>=0;i--)
     {
      addsolution(cluestr[i].dcdata,i);
     }
     if(nsol>solnmaxn)
     {
      check=-1;
      break;
     }
    }
    else
    check=search(clue+1);
   }
  }
  else
  if(firstalpha<firstspace)
  {                   /* have some alpha data to look up */
   check=findwordx1(string,dc,&words,firstspace,len,clue);

   if(nsol>solnmaxn)
   {
    check=-1;
    break;
   }

   if((check>=0) && (check++>8))
   {
    poll();
    check=0;
    if(buttonsevent())
    {
     check=-1;
     break;
    }
   }
   else if(check<0) break;

  }
  else
  {                    /* hard case, got to scan the lot */
   locword(0,dp,&words);
   do
   {
    if(len==(int)strlen(words.string))
    {
     for(i=0;i<len;i++)
     {
      if(string[i]!=' ' && toupper(words.string[i])!=toupper(string[i])) break;
      else
      if(words.string[i] && !string[i]) break;
     }
 
     if(i==len && !string[i])
     {
      cluestr[clue].dcdata=(dcmap[dc]<<24)+words.wordn;

      if(clue==(solwidth-1))
      {
       for(i=clue;i>=0;i--)
       {
        addsolution(cluestr[i].dcdata,i);
       }
       if(nsol>solnmaxn)
       {
        check=-1;
        break;
       }
      }
      else
      {
       fillinstring(clue,words.string);
       check=search(clue+1);
       fillinstring(clue,string);
      }
     }
    }

    if((check>=0) && (check++==2048))
    {
     poll();
     check=0;
     if(buttonsevent())
     {
      check=-1;
      break;
     }
    }
    else if(check<0) break;

   } while(nextword(dp,&words));
   if(check<0) break;
  }
 }

 hourglassoff();

 return(check);
}



int cmpchar(const void * p1,const void * p2)
{
 return((*(char*)p2)-(*(char*)p1));
}


os_error * anagramsearch(char * string)
{
 os_error * err;
 word   words;
 dict * dp;
 int    dc;
 int    len;
 int    i;
 int    j;
 int    ct;
 char   b[STRLEN];
 int    check;

 err=NULL;

 len=strlen(string);

 if(!err)
 {
  qsort(string,len,1,cmpchar);

  check=0;

  hourglasson();

  for(dc=0;dc<nodc;dc++)
  {
   dp=&dtable[dcmap[dc]].dictp;

   if(locword(0,dp,&words))
   {
    do
    {
     if(len==(int)strlen(words.string))
     {
      strcpy(b,string);
      for(ct=0,i=0;i<len;i++)
      {
       if(ct!=i) break;
       for(j=0;j<len;j++)
       {
        if(toupper(b[j])==toupper(words.string[i]) || b[j]=='*')
        {
         b[j]=0;
         ct++;
         break;
        }
       }
      }

      if(i==len && ct==len)
      {
       addsolution((dcmap[dc]<<24)+words.wordn,0);
       if(nsol>solnmaxn)
       {
        check=-1;
        break;
       }
      }
     }

     if((check>=0) && (check++==2048))
     {
      check=0;
      if(buttonsevent())
      {
       check=-1;
       break;
      }
     }
    } while(nextword(dp,&words));
    if(check<0) break;
   }
  }
  hourglassoff();
 }

 return(err);
}



int wordsearch(char * string)
{
 word words;
 int  dc;
 int  flags=0x20000;

 for(dc=0;dc<nodc;dc++)
 {
  if(findword(string,&dtable[dcmap[dc]].dictp,&words,&flags))
  {
   return(1);
  }
 }
 return(0);
}



static int rdc;

void randomwordstart(void)
{
 rdc=0;
}


void randomword(char * string)
{
 int n;
 int r;

 rdc++;
 if(rdc>=nodc) rdc=0;

 n=dtable[dcmap[rdc]].dictp.nowords;

 r=rand()+(rand()<<15);
 r&=0xFFFFFF;

 if(n==0) n=0;
 else     n=r%n;

 strcpy(string,wordn(n,&dtable[dcmap[rdc]].dictp));
}




/* called from main to boot spell check code */

os_error * initsp(void)
{
 inittable();
 loadnames();
 linkmenus();
 return(NULL);
}


void finitsp(void)
{
 unlinkmenus();
}
